start.time  <-  Sys.time()

#########################
# LIBRARY AND FUNCTIONS #
#########################

# Use the following line to load the Snakemake object to manually rerun this script (e.g., for debugging purposes)
# Replace {outputFolder} correspondingly.
# snakemake = readRDS("{outputFolder}/LOGS_AND_BENCHMARKS/2.DESeqPeaks.R.rds")


library("checkmate")
assertClass(snakemake, "Snakemake")
assertDirectoryExists(snakemake@config$additionalInput$scriptsDir)
source(paste0(snakemake@config$additionalInput$scriptsDir, "/functions.R"))

########################################################################
# SAVE SNAKEMAKE S4 OBJECT THAT IS PASSED ALONG FOR DEBUGGING PURPOSES #
########################################################################

createDebugFile(snakemake)

initFunctionsScript(packagesReq = NULL, minRVersion = "3.1.0", warningsLevel = 1, disableScientificNotation = TRUE)
checkAndLoadPackages(c("tidyverse", "futile.logger",  "checkmate", "tools", "methods", "DESeq2"), verbose = FALSE)



###################
#### PARAMETERS ###
###################

par.l = list()

par.l$verbose = TRUE
par.l$log_minlevel = "INFO"

#####################
# VERIFY PARAMETERS #
#####################

checkAndLogWarningsAndErrors(snakemake, checkClass(snakemake, "Snakemake"))

## INPUT ##
checkAndLogWarningsAndErrors(snakemake@input, checkList(snakemake@input, min.len = 2))
checkAndLogWarningsAndErrors(snakemake@input, checkSubset(names(snakemake@input), c("", "metadata", "rawCounts")))


par.l$file_metadata = snakemake@input$metadata
par.l$file_rawCounts = snakemake@input$rawCounts


checkAndLogWarningsAndErrors(par.l$file_metadata, checkFileExists(par.l$file_metadata, access = "r"))
checkAndLogWarningsAndErrors(par.l$file_rawCounts, checkFileExists(par.l$file_rawCounts, access = "r"))


## OUTPUT ##
checkAndLogWarningsAndErrors(snakemake@output, checkList(snakemake@output, min.len = 2))
checkAndLogWarningsAndErrors(names(snakemake@output), checkSubset(names(snakemake@output), c("", "pdf", "data")))

par.l$file_output_pdf       = snakemake@output$pdf
par.l$file_output_data      = snakemake@output$data

## PARAMS ##
checkAndLogWarningsAndErrors(snakemake@params, checkList(snakemake@params, min.len = 3))
#checkAndLogWarningsAndErrors(names(snakemake@params), checkSubset(names(snakemake@params), c("", "binSize", "filterChr", "splitByGCCorrection", "minNoCountsRows")))
checkAndLogWarningsAndErrors(names(snakemake@params), checkSubset(names(snakemake@params), c("", "filterChr", "splitByGCCorrection", "minNoCountsRows")))

#par.l$binSize = as.integer(snakemake@params$binSize)
#checkAndLogWarningsAndErrors(par.l$binSize, checkIntegerish(par.l$binSize))

par.l$filterChr = as.logical(snakemake@params$filterChr)
checkAndLogWarningsAndErrors(par.l$filterChr, checkFlag(par.l$filterChr))

par.l$splitByGCCorrection = as.logical(snakemake@params$splitByGCCorrection)
checkAndLogWarningsAndErrors(par.l$splitByGCCorrection, checkFlag(par.l$splitByGCCorrection))

par.l$minNoCountsRows = as.integer(snakemake@params$minNoCountsRows)
checkAndLogWarningsAndErrors(par.l$minNoCountsRows, checkIntegerish(par.l$minNoCountsRows))

## LOG ##
checkAndLogWarningsAndErrors(snakemake@log, checkList(snakemake@log, min.len = 1))
par.l$file_log = snakemake@log[[1]]


allDirs = c(dirname(par.l$file_output_pdf),
            dirname(par.l$file_log)
)

testExistanceAndCreateDirectoriesRecursively(allDirs)


######################
# FINAL PREPARATIONS #
######################
startLogger(par.l$file_log, par.l$log_minlevel,  removeOldLog = TRUE)
printParametersLog(par.l)


flog.info(paste0("Generate PCA plots"))


# TODO: c("sampleName", "individual") populate automatically

# prepare tables here already

metadata = read_tsv(par.l$file_metadata)

if (nrow(problems(metadata)) > 0) {
  flog.fatal(paste0("Parsing errors: "), problems(metadata), capture = TRUE)
  stop("Error when parsing the file ", par.l$file_metadata, ", see errors above")
}

metadataToInclude = colnames(metadata)
removeCols = c("path_inputForward", "path_inputReverse")
# Check for only NA columns and remove
for (colCur in metadataToInclude) {
  if (all(is.na(metadata[,colCur]))) {
    removeCols = c(removeCols, colCur)
  }
}
metadataToInclude = setdiff(metadataToInclude, removeCols)
# metadata = dplyr::mutate_if(metadata, is.character,as.factor)
metadata = dplyr::mutate(metadata, replicate_No = as.factor(replicate_No))

countsPCA = read_tsv(par.l$file_rawCounts, skip = 1)
if (nrow(problems(countsPCA)) > 0) {
  flog.fatal(paste0("Parsing errors: "), problems(countsPCA), capture = TRUE)
  stop("Error when parsing the file ", par.l$file_rawCounts, ", see errors above")
}

# Test purposes
colnames(countsPCA) = gsub(".final.bam", "", colnames(countsPCA))
colnames(countsPCA) = gsub("'", "", colnames(countsPCA))
colnames(countsPCA) = basename(colnames(countsPCA))



if (par.l$filterChr) {
  countsPCA = dplyr::filter(countsPCA, !grepl("random|chrUn|chrM|hap|_gl", Chr, perl = TRUE))
}

# Keep only the Chr column, as needed for potential filtering by chromosome in the doPCAPlot function
countsPCA = dplyr::select(countsPCA, -Geneid, -Chr, -Start, -End, -Strand, -Length)

# TODO: Loop over all input files or run rule separately for each?


doPCAPlot(countsPCA = countsPCA, 
          metadata = metadata, 
          file_output_PDF = par.l$file_output_pdf, 
          file_output_data = par.l$file_output_data,
          file_log = par.l$file_log, 
          metadataToInclude = metadataToInclude, 
          descriptionString = "variable peaks", 
          splitByGCCorrection = par.l$splitByGCCorrection, 
          minNoCountsRows = par.l$minNoCountsRows)
