start.time  <-  Sys.time()



#########################
# LIBRARY AND FUNCTIONS #
#########################

library("checkmate")
assertClass(snakemake, "Snakemake")
assertDirectoryExists(snakemake@config$additionalInput$scriptsDir)
source(paste0(snakemake@config$additionalInput$scriptsDir, "/functions.R"))

initFunctionsScript(packagesReq = NULL, minRVersion = "3.1.0", warningsLevel = 1, disableScientificNotation = TRUE)
checkAndLoadPackages(c("tidyverse", "futile.logger", "DiffBind", "checkmate", "stats", "stringr"), verbose = FALSE)

########################################################################
# SAVE SNAKEMAKE S4 OBJECT THAT IS PASSED ALONG FOR DEBUGGING PURPOSES #
########################################################################

# Use the following line to load the Snakemake object to manually rerun this script (e.g., for debugging purposes)
# Replace {outputFolder} correspondingly.
# snakemake = readRDS("{outputFolder}/LOGS_AND_BENCHMARKS/1.createConsensusPeaks.R.rds")
createDebugFile(snakemake)

###################
#### PARAMETERS ###
###################

par.l = list()
par.l$allPeakTypes = c("broadPeak", "gappedPeak", "narrowPeak", "stringent", "non-stringent")
par.l$verbose = TRUE
par.l$log_minlevel = "INFO"

#####################
# VERIFY PARAMETERS #
#####################


assertClass(snakemake, "Snakemake")

## INPUT ##
assertList(snakemake@input, min.len = 1)
assertSubset(names(snakemake@input), c("", "stringent", "nonStringent", "encode"))

par.l$files_input_stringent = snakemake@input$stringent
par.l$files_input_nonStringent = snakemake@input$nonStringent
par.l$files_input_ENCODE = snakemake@input$encode
#par.l$file_input_sampleFile = snakemake@input$sampleFile

## OUTPUT ##
assertList(snakemake@output, min.len = 1)
assertSubset(names(snakemake@output), c("", "consensusPeaks_stringent", "consensusPeaks_nonStringent", "consensusPeaks_ENCODE", "summaryPlot", "summaryData"))

par.l$files_output_consensusPeaks_stringent   = snakemake@output$consensusPeaks_stringent
par.l$files_output_consensusPeaks_nonStringent = snakemake@output$consensusPeaks_nonStringent
par.l$files_output_consensusPeaks_ENCODE      = snakemake@output$consensusPeaks_ENCODE
par.l$summaryPlot                              = snakemake@output$summaryPlot
par.l$summaryData                              = snakemake@output$summaryData


## CONFIG ##
assertList(snakemake@config, min.len = 1)




## PARAMS ##
assertList(snakemake@params, min.len = 2)

par.l$GCBiasStr = snakemake@params$GCString
par.l$minOverlap = unique(as.integer(strsplit(snakemake@params$minOverlapValues, ",")[[1]]))

## LOG ##
assertList(snakemake@log, min.len = 1)
par.l$file_log = snakemake@log[[1]]


# allDirs = unique(c(dirname(par.l$files_output_consensusPeaks_stringent), 
#             dirname(par.l$files_output_consensusPeaks_nonStringent),
#             dirname(par.l$files_output_consensusPeaks_ENCODE),
#             dirname(par.l$file_log))
# )

outputs = c(par.l$files_output_consensusPeaks_stringent,
            par.l$files_output_consensusPeaks_nonStringent,
            par.l$files_output_consensusPeaks_ENCODE,
            par.l$file_log)
allDirs = unique(dirname(outputs[!is.null(outputs)]))



testExistanceAndCreateDirectoriesRecursively(allDirs)


######################
# FINAL PREPARATIONS #
######################
startLogger(par.l$file_log, par.l$log_minlevel, removeOldLog = TRUE)
printParametersLog(par.l)

##########################
# Create consensus peaks #
###########################


# Loop over all possible iterations
consensusPeak.stats.df = tibble::tribble(~GCBiasStr, ~peakType, ~minOverlap, ~file, ~nPeaks, ~isError)


for (GCBiasStrCur in par.l$GCBiasStr) {
    
    flog.info(paste0("Calculating consensus peaks for GCBiasStrCur \"", GCBiasStrCur, "\""))
    
    for (peakType in par.l$allPeakTypes) {
        
        flog.info(paste0("Calculating consensus peaks for peak type \"", peakType, "\""))
        
        if (peakType %in% c("broadPeak", "gappedPeak", "narrowPeak")) {
            filesInput  = par.l$files_input_ENCODE
            if (is.null(filesInput)){next}
            filesOutput = par.l$files_output_consensusPeaks_ENCODE
        } else if (peakType == "stringent") {
            filesInput  = par.l$files_input_stringent
            filesOutput = par.l$files_output_consensusPeaks_stringent
        } else if (peakType == "non-stringent") {
            filesInput  = par.l$files_input_nonStringent
            filesOutput = par.l$files_output_consensusPeaks_nonStringent
        } else {
            messageError = paste0("peakType unsupported.")
            flog.error(messageError)
            stop(messageError)
        }
        
        matches = grepl(GCBiasStrCur, filesInput, fixed = TRUE)
        matches2 = grepl(GCBiasStrCur, filesOutput, fixed = TRUE)
        if (GCBiasStrCur != "") {
            filesInput  = filesInput[matches]
            filesOutput = filesOutput[matches2]
        } else {
          if (length(par.l$GCBiasStr) == 1) {
            # Nothing to filter, as only the non-GC bias corrected versions should be produced
          } else {
            # Filter the non-GC bias files out
            matches = grepl(par.l$GCBiasStr[which(par.l$GCBiasStr != "")], filesInput, fixed = TRUE)
            matches2 = grepl(par.l$GCBiasStr[which(par.l$GCBiasStr != "")], filesOutput, fixed = TRUE)
            filesInput  = filesInput[!matches]
            filesOutput = filesOutput[!matches2]
          }
            
        }
        
        # Differentiate the files for Encode now, as there are 3 different flavors for ENCODE
        if (peakType %in% c("broadPeak", "gappedPeak", "narrowPeak")) {
            filesInput  = str_subset(filesInput, peakType)
            filesOutput = str_subset(filesOutput, peakType)
        }
        
  

        sampleMetaData.df = tibble(Peaks = filesInput, PeakCaller = "narrow")


        assertIntegerish(max(par.l$minOverlap), upper = nrow(sampleMetaData.df), lower = 1)
        
        # Check if there are any peak fiels with 0 lines. diffBind crashes upon this otherwise
        for (rowCur in 1:nrow(sampleMetaData.df)) {
            
            # readr > 1.1.1 and <=1.3.1 encounters a bug with empty gz files
            peaksCur <- tryCatch({
                read_tsv(sampleMetaData.df$Peaks[rowCur], col_names = FALSE, col_types = cols())
                
            }, error = function(err) {
                
                tibble()
            })
            
            flog.info(paste0("Found ", nrow(peaksCur), " peaks in file ", sampleMetaData.df$Peaks[rowCur]))
            if (nrow(peaksCur) == 0) {
                sampleMetaData.df$Peaks[rowCur] = NA
                flog.warn(paste0(" Skip file for consensus peaks generation "))
            }
            
           
        }
        
        # DiffBind assigning. Convert to regular data frame first so DiffBind does not complain
        dba <- dba(sampleSheet = as.data.frame(sampleMetaData.df))
        
        
        for (minOverlapCur in par.l$minOverlap) {
            
            flog.info(paste0("  Calculating consensus peaks for minOverlap ", minOverlapCur, " for ", peakType, " peaks..."))
            
            # Calculate consensus peaks
            consensusPeaks.df = tryCatch({
                
                consensus =  dba.peakset(dba, minOverlap = minOverlapCur)
                
                # Get the consensus peaks (last row) and modify
                
                consensusPeaks.df = consensus$peaks[[nrow(sampleMetaData.df) + 1]] %>%
                    as_tibble() %>%
                    dplyr::rename(chr = V1, start = V2, end = V3, pvalue = V4) %>%
                    dplyr::mutate(length = end - start,
                           annotation = paste0(chr, ":", start, "-", end),
                           strand = ".") %>%
                    dplyr::select(chr, start, end, annotation, pvalue, strand)
                
                
                consensusPeaks.df
                
            }, error = function(e) {
                
                consensusPeaks.df = tribble(~chr, ~start, ~end, ~annotation, ~pvalue, ~strand)
                consensusPeaks.df
            })
           

            outputCur = str_subset(filesOutput, paste0("minOverlap", minOverlapCur, ".bed.gz"))
            stopifnot(length(outputCur) == 1)
            
            consensusPeak.stats.df = add_row(consensusPeak.stats.df,
                                             GCBiasStr = GCBiasStrCur,
                                             peakType = peakType,
                                             minOverlap = minOverlapCur,
                                             file = outputCur,
                                             nPeaks = nrow(consensusPeaks.df),
                                             isError = nrow(consensusPeaks.df) == 0)
            
            if (nrow(consensusPeaks.df) > 0) {
                consensusPeaks.df = mutate_if(consensusPeaks.df, is.numeric, as.character)
                write_tsv(consensusPeaks.df, outputCur, col_names = FALSE)

            } else {
                write_lines("# No consensus peaks found", outputCur)
            }
            
        }
        
        
        
    }
}


# Save consensusPeak.stats.df and plot
pdf(par.l$summaryPlot)
ggplot(consensusPeak.stats.df, aes(minOverlap, nPeaks, color = peakType)) + geom_line() + geom_point() + scale_color_viridis_d() +
  theme_bw() + ggtitle("Number of consensus peaks overview")

ggplot(consensusPeak.stats.df, aes(minOverlap, nPeaks, color = peakType)) + geom_line() + geom_point() + scale_color_viridis_d() +
  theme_bw() + ggtitle("Number of consensus peaks overview (y capped to 200,000") + ylim(0,200000)

dev.off()

write_tsv(consensusPeak.stats.df, par.l$summaryData)


.printExecutionTime(start.time)

flog.info("Session info: ", sessionInfo(), capture = TRUE)

